#include <stdio.h>      // standard input / output functions
#include <stdlib.h>
#include <string.h>     // string function definitions
//#include <strings.h>
#include <unistd.h>     // UNIX standard function definitions
#include <sys/ioctl.h>
#include <fcntl.h>      // File control definitions
#include <errno.h>      // Error number definitions
#include <termios.h>
//#include <asm/termios.h>    // POSIX terminal control definitions
#include <time.h>
#include <math.h>
#define MAXOPCIONES 4
/**********************************************************
 conversión de unidades
 **********************************************************/
// datos en semipasos, resultado en ångströms (*5/2)
#define semipasos2lambda(p) ((p*5)/2)
// datos en ångströms, resultado en semipasos (*2/5) (*8/20)
#define lambda2semipasos(l) ((l*2)/5)
// datos en ångströms, resultado en pasos (/5) (*4/20)
#define lambda2pasos(l) (l/5)

// Variables globales:
char opciones[MAXOPCIONES][80] = { "Salir.", "Ir a una longitud de onda.",
		"Seleccionar ganancia del fotodiodo.", "Hacer un barrido." };
int posActualAng = 0, mul = 1, cero = 0;
double posActual = 0, posActualNm = 0;
// Variable globales para la utilización del puerto USB para comunicarnos con el micro.
int fd, n; /* File descriptor for the port */
struct termios term;
FILE *fp;
static char buf[256];

void go2lambda(double posNuevaAng) {
	int semiPasos = 0;
	double angstroms = 0;
	char str[6];

	angstroms = posNuevaAng - posActualAng;
	semiPasos = lambda2semipasos(angstroms);
	if (semiPasos < 0) {
		// Retrocedemos.
		semiPasos = 0 - semiPasos;
		sprintf(str, "R%d\n", semiPasos); // Formamos la cadena str
		// Escribir por el puerto serie.
		n = write(fd, str, strlen(str));
		if (n < 0)
			fputs("write() of 4 bytes failed!\n", stderr);
		// Leer por el puerto serie.
		fgets(buf, 255, fp); // Leo.
		//printf(">%s<\n", buf); // Muestro.
		// Comprobamos que el micro nos ha contestado con OK.
		sprintf(str, "R%dOK\n", semiPasos);
		if (strcmp(str, buf) != 0) {
			printf(
					"\n ATENCIÓN\n ¡¡Hemos llegado al final de carrera por abajo!!\n");
			posNuevaAng = -208; // Long. de onda negativa, no tiene sentido, el espejo se ha pasado del límite.
		}
	} else {
		// Avanzamos
		sprintf(str, "A%d\n", semiPasos); // Formamos la cadena str
		// Escribir por el puerto serie.
		n = write(fd, str, strlen(str));
		if (n < 0)
			fputs("write() of 4 bytes failed!\n", stderr);
		// Leer por el puerto serie.
		fgets(buf, 255, fp); // Leo.
		//printf(">%s<\n", buf); // Muestro.
		// Comprobamos que el micro nos ha contestado con OK.
		sprintf(str, "A%dOK\n", semiPasos);
		if (strcmp(str, buf) != 0) {
			printf(
					"\n ATENCIÓN\n ¡¡Hemos llegado al final de carrera por arriba!!\n");
			posNuevaAng = 18676;
		}
	}
	posActualAng = posNuevaAng;
	posActualNm = (double) (1.0 * posActualAng / 10.0);
	posActual = posActualNm / 2;
	return;
}

void SelectorGanancia() {
	int opcion = 0, opcValida = 0;
	char str[4];
	while (opcValida == 0) {
		system("clear");
		printf(
				"\n\n/=============================================================================/\n");
		printf(
				"/                         FUNCIÓN SELECTOR DE GANANCIA                        /\n");
		printf(
				"/=============================================================================/\n");

		printf("\r\n\nSeleccionar ganancia:\r\n");
		printf("\r===========================================\r\n");
		printf("\r1.- Ganancia x1.\r\n");
		printf("\r2.- Ganancia x4.\r\n");
		printf("\r3.- Ganancia x16.\r\n");
		printf("\r4.- Ganancia x64.\r\n");
		printf("\r===========================================\r\n\n");

		printf("\r\n\nEscoja una opción: \r\n\n");
		scanf("%d", &opcion);
		fflush(stdin);

		switch (opcion) {
		case 1: // 49=1 en ASCII
			opcValida = 1;
			printf("\r\n\nGanancia seleccionada: x1.\r\n\n");
			// Pasamos al micro el comando: G1
			strcpy(str, "G1\n"); // Formamos la cadena str
			mul = 1;
			cero = 0;
			// Escribir por el puerto serie.
			n = write(fd, str, strlen(str));
			if (n < 0)
				fputs("write() of 3 bytes failed!\n", stderr);
			// Leer por el puerto serie.
			fgets(buf, 255, fp); // Leo.
			//printf(">%s<\n", buf); // Muestro.
			break;
		case 2: // 50=2 en ASCII
			opcValida = 1;
			printf("\r\n\nGanancia seleccionada: x4.\r\n\n");
			// Pasamos al micro el comando: G2
			strcpy(str, "G2\n"); // Formamos la cadena str
			mul = 4;
			cero = 2;
			// Escribir por el puerto serie.
			n = write(fd, str, strlen(str));
			if (n < 0)
				fputs("write() of 3 bytes failed!\n", stderr);
			// Leer por el puerto serie.
			fgets(buf, 255, fp); // Leo.
			//printf(">%s<\n", buf); // Muestro.
			break;
		case 3: // 51=3 en ASCII
			opcValida = 1;
			printf("\r\n\nGanancia seleccionada: x16.\r\n\n");
			// Pasamos al micro el comando: G3
			strcpy(str, "G3\n"); // Formamos la cadena str
			mul = 16;
			cero = 24;
			// Escribir por el puerto serie.
			n = write(fd, str, strlen(str));
			if (n < 0)
				fputs("write() of 3 bytes failed!\n", stderr);
			// Leer por el puerto serie.
			fgets(buf, 255, fp); // Leo.
			//printf(">%s<\n", buf); // Muestro.
			break;
		case 4: // 52=4 en ASCII
			opcValida = 1;
			printf("\r\n\nGanancia seleccionada: x64.\r\n\n");
			// Pasamos al micro el comando: G4
			strcpy(str, "G4\n"); // Formamos la cadena str
			mul = 64;
			cero = 115;
			// Escribir por el puerto serie.
			n = write(fd, str, strlen(str));
			if (n < 0)
				fputs("write() of 3 bytes failed!\n", stderr);
			// Leer por el puerto serie.
			fgets(buf, 255, fp); // Leo.
			//printf(">%s<\n", buf); // Muestro.
			break;
		default:
			opcValida = 0;
			printf("\r\n\nLa opción seleccionada no es valida.");
			printf("\r\nPor favor, vuelva a intentarlo.\r\n\n");
			printf("Pulse enter para continuar:");
			getchar();
			getchar();
			break;
		}
	}
	printf("Pulse enter para continuar:");
	getchar();
	getchar();
	system("clear");
	return;
}

void IrPosicion() {
	int posNuevaAng = 0, i = 0, j = 1, medida = 0;
	char str[6];

	system("clear");
	printf(
			"\n\n/=============================================================================/\n");
	printf(
			"/                      FUNCIÓN IR A UNA LONGITUD DE ONDA                      /\n");
	printf(
			"/=============================================================================/\n");
	printf("\nLa posición actual del monocromador es %.1f.\n", posActual);
	printf("\nEquivale a: %.1f nm.\n", posActualNm);
	printf("\n            %d Å.\n", posActualAng);
	printf(
			"\nIntroduzca la longitud de onda a la que desea ir (en ångströms):");
	scanf("%d", &posNuevaAng);
	fflush(stdin);

	printf("\nMoviendo motor hacia long. de onda %d Å.\n", posNuevaAng);
	printf("===========================================\n");

	go2lambda(posNuevaAng); // Movemos a long. de onda.

	// Encendemos LED.
	strcpy(str, "E\n"); // Formamos la cadena str
	// Escribir por el puerto serie.
	n = write(fd, str, strlen(str));
	if (n < 0)
		fputs("write() of 3 bytes failed!\n", stderr);
	// Leer por el puerto serie.
	fgets(buf, 255, fp); // Leo.
	//printf(">%s<\n", buf); // Muestro.

	// Medimos.
	strcpy(str, "M\n"); // Formamos la cadena str
	// Escribir por el puerto serie.
	n = write(fd, str, strlen(str));
	if (n < 0)
		fputs("write() of 3 bytes failed!\n", stderr);
	// Leer por el puerto serie.
	fgets(buf, 255, fp); // Leo.
	//printf(">%s<\n", buf); // Muestro.
	for (i = strlen(buf) - 2; i >= 1; i--) {
		medida += (buf[i] - 48) * j;
		j *= 10;
	}

	// Apagamos LED.
	strcpy(str, "O\n"); // Formamos la cadena str
	// Escribir por el puerto serie.
	n = write(fd, str, strlen(str));
	if (n < 0)
		fputs("write() of 3 bytes failed!\n", stderr);
	// Leer por el puerto serie.
	fgets(buf, 255, fp); // Leo.
	//printf(">%s<\n", buf); // Muestro.

	printf("\nLa medida obtenida es: %d\n\n", medida);
	printf("Pulse enter para continuar:");
	getchar();
	getchar();
	return;
}

void Barrido() {
	double posInic = 0, posFin = 0, incremento = 0, lmin = 0, X = 0, Xcorr = 0,
			F = 0;
	double F2a1 = -2.23966381310854, F2b1 = 527.67116064641, F2c1 =
			-244.889472096371, F2a2 = 1.43557861825581, F2b2 = 1041.73978781212,
			F2c2 = -42.5977692225163, F2a3 = 3.59531261238112e-2, F2b3 =
					911.490769664776, F2c3 = 14.3208742339473;
	int opcion = 0, opcValida = 0, medida = 0, i = 0, j = 1, Y = 0, Ycorr = 0,
			maximo[6] = { 0, 0, 0, 0, 0, 0 };
	char nombreFichero[255], str[6], cambia;
	FILE *fichero, *pipe;
	time_t tiempo = time(0);
	system("clear");
	printf(
			"\n\n/=============================================================================/\n");
	printf(
			"/                           FUNCIÓN HACER UN ABRRIDO                          /\n");
	printf(
			"/=============================================================================/\n");
	printf("\nLa posición actual del monocromador es %.1f.\n", posActual);
	printf("\nEquivale a: %.1f nm.\n", posActualNm);
	printf("\n            %d Å.\n", posActualAng);
	printf("\nIntroduza posición inicial (en ångströms):");
	scanf("%lf", &posInic);
	fflush(stdin);
	lmin = posInic;

	printf("\nIntroduza posición final (en ångströms):");
	scanf("%lf", &posFin);
	fflush(stdin);

	while (opcValida == 0) {
		system("clear");
		printf(
				"\n\n/=============================================================================/\n");
		printf(
				"/                            SELECTOR DE RESOLUCIÓN                           /\n");
		printf(
				"/=============================================================================/\n");

		printf("\n\nSeleccionar resolución:\n");
		printf("===========================================\n");
		printf("1.- 2.5 A\t1 semipaso.\n");
		printf("2.- 5.0 A\t2 semipasos.\n");
		printf("3.- 10 A\t4 semipasos.\n");
		printf("4.- 20 A\t8 semipasos.\n");
		printf("===========================================\n\n");

		printf("\r\n\nEscoja una opción: \r\n\n");
		scanf("%d", &opcion);
		fflush(stdin);

		switch (opcion) {
		case 1: // 49=1 en ASCII
			opcValida = 1;
			printf("\n\nResolución seleccionada: 2.5 Å\t1 semipaso.\n\n");
			incremento = 1;
			break;
		case 2: // 50=2 en ASCII
			opcValida = 1;
			printf("\n\nResolución seleccionada: 5 Å\t2 semipaso.\n\n");
			incremento = 2;
			break;
		case 3: // 51=3 en ASCII
			opcValida = 1;
			printf("\n\nResolución seleccionada: 10 Å\t4 semipaso.\n\n");
			incremento = 4;
			break;
		case 4: // 52=4 en ASCII
			opcValida = 1;
			printf("\n\nResolución seleccionada: 20 Å\t8 semipaso.\n\n");
			incremento = 8;
			break;
		default:
			opcValida = 0;
			printf("\n\nLa opción seleccionada no es valida.");
			printf("\nPor favor, vuelva a intentarlo.\n\n");
			printf("Pulse enter para continuar:");
			getchar();
			getchar();
			break;
		}
	}

	// INCLUIMOS EL SELECTOR DE GANACIA
	SelectorGanancia();

	system("clear");
	printf(
			"\n\n/=============================================================================/\n");
	printf(
			"/                           FUNCIÓN HACER UN ABRRIDO                          /\n");
	printf(
			"/=============================================================================/\n");
	printf("\nLa posición actual del monocromador es %.1f.\n", posActual);
	printf("\nEquivale a: %.1f nm.\n", posActualNm);
	printf("\n            %d Å.\n", posActualAng);
	printf("\nDatos del barrido:\n");
	printf(
			"/-----------------------------------------------------------------------------/\n");
	printf("\nLa posición incial es: %.1f Å", posInic);
	printf("\nLa posición final es: %.1f Å", posFin);
	printf("\nEl incremento es: %.1f Å", semipasos2lambda(incremento));
	printf("\nLa ganancia seleccionada es: %d", mul);
	printf(
			"\n/-----------------------------------------------------------------------------/\n");

	struct tm *tlocal = localtime(&tiempo);
	char output[128];
	strftime(output, 128, "%y%m%d%H%M%S", tlocal);

	// Preguntamos al usuario si desea cambiar el nombre por defecto del fichero en el que se guardan los datos.
	printf(
			"\n\nPor defecto, el nombre del fichero en el que se van a almacenar los datos es:\n\t\tB20%s-i%.1f-f%.1f-R%.1f-Gx%d",
			output, posInic / 10, posFin / 10,
			semipasos2lambda(incremento) / 10, mul);
	printf("\n¿Desea cambiar el nombre del fichero? (S/N):");
	scanf("%c", &cambia);
	fflush(stdin);
	if (cambia == 'S' || cambia == 's') { //Recogemos el nuevo nombre del fichero.
		printf("\nIntroduzca el nuevo nombre del fichero: ");
		scanf("%s", nombreFichero);
		fflush(stdin);
		printf("\nEl nombre del fichero es: %s.", nombreFichero);
	} else
		sprintf(nombreFichero, "B20%s-i%.1f-f%.1f-R%.1f-Gx%d", output,
				posInic / 10, posFin / 10, semipasos2lambda(incremento) / 10,
				mul);

	fichero = fopen(nombreFichero, "w"); //Abrimos fichero.

	printf("\nRealizando barrido.\n");
	printf("===========================================\n");

	// Encendemos LED.
	strcpy(str, "E\n"); // Formamos la cadena str
	// Escribir por el puerto serie.
	n = write(fd, str, strlen(str));
	if (n < 0)
		fputs("write() of 3 bytes failed!\n", stderr);
	// Leer por el puerto serie.
	fgets(buf, 255, fp); // Leo.
	//printf(">%s<\n", buf); // Muestro.

	while (posInic <= posFin) {
		go2lambda(posInic); // Movemos a long. de onda.
		// Medimos.
		strcpy(str, "M\n"); // Formamos la cadena str.
		// Escribir por el puerto serie.
		n = write(fd, str, strlen(str));
		if (n < 0)
			fputs("write() of 3 bytes failed!\n", stderr);
		// Leer por el puerto serie.
		fgets(buf, 255, fp); // Leo.
		//printf(">%s<\n", buf); // Muestro.
		for (i = strlen(buf) - 2; i >= 1; i--) {
			medida += (buf[i] - 48) * j;
			j *= 10;
		}
		// Calculamos X, Y y las corregidas (menos el cero).
		Y = medida;
		Ycorr = Y - cero;
		F = F2a1 / (1 + exp(-((posInic / 10.0) - F2b1) / F2c1))
				+ F2a2 / (1 + exp(-((posInic / 10.0) - F2b2) / F2c2))
				+ F2a3 / (1 + exp(-((posInic / 10.0) - F2b3) / F2c3));
		X = Y / F;
		Xcorr = Ycorr / F;
		// Almacenamos los máximos.
		if (Ycorr > maximo[2])
			maximo[2] = Ycorr;
		if (Y > maximo[3])
			maximo[3] = Y;
		if (Xcorr > maximo[4])
			maximo[4] = Xcorr;
		if (X > maximo[5])
			maximo[5] = X;
		fprintf(fichero, "\t%.1f\t%d\t%d\t%.4f\t%.4f\t%.4f\n", posInic, Ycorr,
				Y, Xcorr, X, F); // Escribimos en el fichero
		printf("\t%.1f\t%d\n", posInic, medida); // Escribimos por pantalla
		fflush(stdout);
		fflush(fichero);
		medida = 0; // Inicializamos medida y j para la siguiente medida.
		j = 1;
		posInic += semipasos2lambda(incremento);
	}
	// Apagamos LED.
	strcpy(str, "O\n"); // Formamos la cadena str
	// Escribir por el puerto serie.
	n = write(fd, str, strlen(str));
	if (n < 0)
		fputs("write() of 3 bytes failed!\n", stderr);
	// Leer por el puerto serie.
	fgets(buf, 255, fp); // Leo.
	//printf(">%s<\n", buf); // Muestro.

	fclose(fichero); // Cerramos el fichero.
	// Code for gnuplot pipe
	pipe = popen("gnuplot -persist", "w");
	fprintf(pipe, "\n");
	// Your code goes here
	fprintf(pipe, "set xrange[%lf:%lf]\n", lmin / 10, posFin / 10);
	fprintf(pipe, "set xlabel \"Longitud de onda (nm)\" \n");
	fprintf(pipe, "set ylabel \"Intensidad (u.a.)\" \n");
	fprintf(pipe, "set title \"Xcorr LED - Res. %.1f nm - PD x%d\" \n",
			semipasos2lambda(incremento) / 10, mul);
	fprintf(pipe, "unset key\n");
	// Pintamos ejex: columna 1 del fichero/10 y ejey: columna 2, with line/point
	//fprintf(pipe, "plot \"%s\" using ($1/10):2 w l\n", nombreFichero);
	// Pintamos ejex: columna 1 del fichero/10 y ejey: columna 4/maximo[4], with line/point
	fprintf(pipe, "set mxtics 5\n");
	fprintf(pipe, "set mytics 5\n");
	//fprintf(pipe, "set grid\n");
	fprintf(pipe, "set grid mxtics mytics xtics ytics\n");
	//fprintf(pipe, "plot \"%s\" using ($1/10):($4/\"%d\") w l\n", nombreFichero, maximo[4]);
	fprintf(pipe, "plot \"%s\" using ($1/10):($4/\"%d\") w l\n", nombreFichero,
			maximo[4]);
	fprintf(pipe, "set terminal postscript color enhanced\n");
	fprintf(pipe, "set output \"%s.eps\"\n", nombreFichero);
	fprintf(pipe, "rep\n");
	// Don't forget to close the pipe
	fclose(pipe);

	printf("\n\nPulse enter para continuar:");
	getchar();
	return;
}

void muestraMenu() {
	int i = 0, opcion = 1;

	while (opcion != 0) {
		system("clear");
		printf(
				"\n\n/=============================================================================/\n");
		printf(
				"/                                MENÚ PRINCIPAL                               /\n");
		printf(
				"/=============================================================================/\n");
		printf("\nLa posición actual del monocromador es %.1f.\n", posActual);
		printf("\nEquivale a: %.1f nm.\n", posActualNm);
		printf("\n            %d Å.\n", posActualAng);
		printf("\n\nMenú principal:\n");
		printf("===========================================\n");
		for (i = 0; i < MAXOPCIONES; i++)
			printf("%d %s\n", i, opciones[i]);

		printf("===========================================\n");
		printf("\nEscoja una opción: ");
		scanf("%d", &opcion);
		fflush(stdin);

		switch (opcion) {
		case 0:
			system("clear");
			printf("\nCerrando aplicación...\n\n");
			printf("Pulse enter para continuar:");
			getchar();
			getchar();
			return;
		case 1:
			IrPosicion();
			break;
		case 2:
			SelectorGanancia();
			break;
		case 3:
			Barrido();
			break;
		default:
			printf("\r\n\nLa opción seleccionada no es valida.");
			printf("\r\nPor favor, vuelva a intentarlo.\r\n\n");
			printf("Pulse enter para continuar:");
			getchar();
			getchar();
			break;
		}
	}
}

int main() {
	system("clear");
	printf(
			"\n\n/=============================================================================/\n");
	printf(
			"/                           PROGRAMA CONTROLADOR DE                           /\n");
	printf(
			"/                              MONOCROMADOR v1.0                              /\n");
	printf(
			"/=============================================================================/\n");
	printf(
			"/                            Jorge Fernández Lucas                            /\n");
	printf(
			"/                          UNIVERSIDAD DE VALLADOLID                          /\n");
	printf(
			"/=============================================================================/\n");

	// Antes de pedir la posición actual tenemos que inicializar el micro y el motor.
	// Mantener al micro a la espera de comando.
	fd = open("/dev/ttyUSB0", O_RDWR | O_NOCTTY); // Abrimos el puerto USB.
	if (fd == -1) { // Comprobamos si lo hemos abierto correctamente.
		perror("open_port: Unable to open /dev/ttyUSB0 - ");
		exit(1);
	}
	if ((fp = fdopen(fd, "rw")) == NULL) { // Al utilizar fdopen podremos utilizar después funciones como fgets
		perror("fdopen"); // muy útil para leer cadenas de caracteres hasta el \n.
		exit(1);
	}
	// Inicializamos los flags.
	tcflush(fd, TCIOFLUSH);
	bzero(&term, sizeof(term));
	term.c_cflag = B115200; // Velocidad del dispositivo.
	term.c_cflag |= CS8 | CLOCAL | CREAD;
	term.c_iflag = IGNPAR | IGNBRK | IXON | IXOFF;
	term.c_oflag = 0;

	/* set input mode (non-canonical, no echo,...) */
	term.c_lflag = 0;

	cfmakeraw(&term); // flags for RAW serial mode

	if (tcsetattr(fd, TCSANOW, &term)) {
		fprintf(stderr, "Fallo en tcsetattr\n");
		exit(1);
	}

	// reset y bootloader H
	if (ioctl(fd, TIOCMGET, &n)) {
		fprintf(stderr, "Fallo en ioctl TIOCMGET\n");
		exit(0);
	}
	n &= ~(TIOCM_DTR | TIOCM_RTS); // Reset y Bootloader H
	if (ioctl(fd, TIOCMSET, &n)) {
		fprintf(stderr, "Fallo en ioctl TIOCMSET\n");
		exit(0);
	}

	//Comprobamos el final de carrera por arriba al realizar el paso inicial del motor.
	// Leer por el puerto serie.
	fgets(buf, 255, fp); // Leo.
	//printf(">%s<\n", buf); // Muestro.
	// Comprobamos que el micro nos ha contestado con OK.
	if (strcmp("fcA\n", buf) == 0)
		printf(
				"\n ATENCIÓN\n ¡¡Hemos llegado al final de carrera por arriba!!\n");

	printf("\n\n\n\nIntroduzca la posición actual del monocromador: ");
	// Así sería para imprimir acentos en windows.
	//printf("\n\n\n\nIntroduzca la posici%cn actual del monocromador: ", 162);
	scanf("%lf", &posActual);
	fflush(stdin);
	posActualNm = 2 * posActual;
	posActualAng = 10 * posActualNm;
	muestraMenu();

	return 0;
}
